declare module 'dojo-app/IdentityRegistry' {
	import { Handle } from 'dojo-core/interfaces';
	/**
	 * Registry identities can be strings or symbols. Note that the empty string is allowed.
	 */
	export type Identity = string | symbol;
	/**
	 * A registry of values, mapped by identities.
	 */
	export default class IdentityRegistry<V extends Object> {
	    private _entryMap;
	    private _idMap;
	    constructor();
	    /**
	     * Look up a value by its identifier.
	     *
	     * Throws if no value has been registered for the given identifier.
	     *
	     * @param id The identifier
	     * @return The value
	     */
	    get(id: Identity): V;
	    /**
	     * Determine whether the value has been registered.
	     * @param value The value
	     * @return `true` if the value has been registered, `false` otherwise
	     */
	    contains(value: V): boolean;
	    /**
	     * Remove from the registry the value for a given identifier.
	     * @param id The identifier
	     * @return `true` if the value was removed, `false` otherwise
	     */
	    delete(id: Identity): boolean;
	    /**
	     * Determine whether a value has been registered for the given identifier.
	     * @param id The identifier
	     * @return `true` if a value has been registered, `false` otherwise
	     */
	    hasId(id: Identity): boolean;
	    /**
	     * Look up the identifier for which the given value has been registered.
	     *
	     * Throws if the value hasn't been registered.
	     *
	     * @param value The value
	     * @return The identifier otherwise
	     */
	    identify(value: V): Identity;
	    /**
	     * Register a new value with a new identity.
	     *
	     * Throws if a different value has already been registered for the given identity,
	     * or if the value has already been registered with a different identity.
	     *
	     * @param id The identifier
	     * @param value The value
	     * @return A handle for deregistering the value. Note that when called repeatedly with
	     *   the same identifier and value combination, the same handle is returned
	     */
	    register(id: Identity, value: V): Handle;
	}

}
declare module 'dojo-app/_moduleResolver' {
	import Promise from 'dojo-shim/Promise';
	/**
	 * Internal interface to asynchronously resolve a module by its identifier.
	 */
	export interface ResolveMid {
	    (mid: string): Promise<any>;
	}
	/**
	 * Function that maps a (relative) module identifier to an absolute one.
	 */
	export interface ToAbsMid {
	    (moduleId: string): string;
	}
	export default function makeResolver(toAbsMid: ToAbsMid): ResolveMid;

}
declare module 'dojo-app/_resolveListenersMap' {
	import { EventedListenersMap } from 'dojo-compose/mixins/createEvented';
	import Promise from 'dojo-shim/Promise';
	import { CombinedRegistry, WidgetListenersMap } from 'dojo-app/createApp';
	export default function resolveListenersMap(registry: CombinedRegistry, listeners: WidgetListenersMap): Promise<EventedListenersMap>;

}
declare module 'dojo-app/_realizeCustomElements' {
	import { Handle } from 'dojo-core/interfaces';
	import Promise from 'dojo-shim/Promise';
	import { CombinedRegistry, StoreLike, WidgetLike } from 'dojo-app/createApp';
	export interface RealizationHandle extends Handle {
	    /**
	     * Get a realized widget with the given ID.
	     * @return The widget, or `null` if no such widget exists
	     */
	    getWidget(id: string): WidgetLike;
	}
	export function isValidName(name: string): boolean;
	export function normalizeName(name: string): string;
	export default function realizeCustomElements(registry: CombinedRegistry, defaultStore: StoreLike, root: Element): Promise<Handle>;

}
declare module 'dojo-app/createApp' {
	import { Action } from 'dojo-actions/createAction';
	import { ComposeFactory } from 'dojo-compose/compose';
	import { Destroyable } from 'dojo-compose/mixins/createDestroyable';
	import { EventedListener } from 'dojo-compose/mixins/createEvented';
	import { ObservableState, State } from 'dojo-compose/mixins/createStateful';
	import { Handle } from 'dojo-core/interfaces';
	import Promise from 'dojo-shim/Promise';
	import { RenderableMixin } from 'dojo-widgets/mixins/createRenderable';
	import { ToAbsMid, ResolveMid } from 'dojo-app/_moduleResolver';
	import { RealizationHandle } from 'dojo-app/_realizeCustomElements';
	export { ToAbsMid };
	/**
	 * Any kind of action.
	 */
	export type ActionLike = Action<any, any, any>;
	/**
	 * Any kind of store.
	 */
	export type StoreLike = ObservableState<State>;
	/**
	 * Any kind of widget.
	 */
	export type WidgetLike = Destroyable & RenderableMixin;
	/**
	 * Factory method to (asynchronously) create an action.
	 *
	 * @param registry The combined registries of the app
	 * @return The action, or a promise for it
	 */
	export interface ActionFactory {
	    (registry: CombinedRegistry): ActionLike | Promise<ActionLike>;
	}
	/**
	 * Factory method to (asynchronously) create a store.
	 *
	 * @return The store, or a promise for it
	 */
	export interface StoreFactory {
	    (options?: Object): StoreLike | Promise<StoreLike>;
	}
	/**
	 * Factory method to (asynchronously) create a widget.
	 *
	 * @return The widget, or a promise for it
	 */
	export interface WidgetFactory {
	    (options?: Object): WidgetLike | Promise<WidgetLike>;
	}
	/**
	 * Plain old JavaScript object that contains definitions of actions, stores and widgets.
	 */
	export interface Definitions {
	    /**
	     * Action definitions.
	     */
	    actions?: ActionDefinition[];
	    /**
	     * Custom element definitions.
	     */
	    customElements?: CustomElementDefinition[];
	    /**
	     * Store definitions.
	     */
	    stores?: StoreDefinition[];
	    /**
	     * Widget definitions.
	     */
	    widgets?: WidgetDefinition[];
	}
	/**
	 * Actions, stores and widgets should have string identifiers.
	 */
	export type Identifier = string;
	/**
	 * Base definition for a single action, store or widget.
	 */
	export interface ItemDefinition<Factory, Instance> {
	    /**
	     * Identifier for which the action, store or widget is to be registered.
	     */
	    id: Identifier;
	    /**
	     * Factory to create an action, store or widget, or a module identifier that resolves to
	     * such a factory.
	     */
	    factory?: Factory | string;
	    /**
	     * An action, store or widget instance, or a module identifier that resolves to such
	     * an instance.
	     */
	    instance?: Instance | string;
	}
	/**
	 * Definition for a single action.
	 */
	export interface ActionDefinition extends ItemDefinition<ActionFactory, ActionLike> {
	    /**
	     * Identifier of a store which the action should observe for its state.
	     *
	     * When the action is created it'll automatically observe this store.
	     */
	    stateFrom?: Identifier | StoreLike;
	}
	/**
	 * Definition for a custom element;
	 */
	export interface CustomElementDefinition {
	    /**
	     * The name of the custom element. Must be valid according to
	     * <https://www.w3.org/TR/custom-elements/#valid-custom-element-name>.
	     */
	    name: string;
	    /**
	     * Factory to create a widget, or a module identifier that resolves to such a factory.
	     */
	    factory: WidgetFactory | string;
	}
	/**
	 * Definition for a single store.
	 */
	export interface StoreDefinition extends ItemDefinition<StoreFactory, StoreLike> {
	    /**
	     * Optional options object passed to the store factory.
	     */
	    options?: Object;
	}
	/**
	 * Definition for a single widget.
	 */
	export interface WidgetDefinition extends ItemDefinition<WidgetFactory, WidgetLike> {
	    /**
	     * Any listeners that should automatically be attached to the widget.
	     */
	    listeners?: WidgetListenersMap;
	    /**
	     * Identifier of a store which the widget should observe for its state.
	     *
	     * When the widget is created, the store is passed as the `stateFrom` option.
	     */
	    stateFrom?: Identifier | StoreLike;
	    /**
	     * Optional options object passed to the widget factory. Must not contain `id` and `stateFrom` properties.
	     */
	    options?: Object;
	}
	/**
	 * A listener for widgets, as used in definitions. May be an identifier for an action or an actual event listener.
	 */
	export type WidgetListener = Identifier | EventedListener<any>;
	export type WidgetListenerOrArray = WidgetListener | WidgetListener[];
	/**
	 * A map of listeners where the key is the event type.
	 */
	export interface WidgetListenersMap {
	    [eventType: string]: WidgetListenerOrArray;
	}
	/**
	 * Read-only interface for the combined registries of the app factory.
	 */
	export interface CombinedRegistry {
	    /**
	     * Get the action with the given identifier.
	     *
	     * Note that the action may still need to be loaded when this method is called.
	     *
	     * @param id Identifier for the action
	     * @return A promise for when the action has been loaded. Rejected if loading fails or if no action is registered
	     *   with the given identifier.
	     */
	    getAction(id: Identifier): Promise<ActionLike>;
	    /**
	     * Check whether an action has been registered with the given identifier.
	     *
	     * @param id Identifier for the action
	     * @return `true` if an action has been registered, `false` otherwise.
	     */
	    hasAction(id: Identifier): boolean;
	    /**
	     * Get the factory for the custom element with the given name.
	     *
	     * @param name Name of the custom element
	     * @return A factory to create a widget for the custom element.
	     */
	    getCustomElementFactory(name: string): WidgetFactory;
	    /**
	     * Check whether a custom element has been registered with the given name.
	     *
	     * @param name Name of the custom element
	     * @return `true` if a custom element has been registered, `false` otherwise.
	     */
	    hasCustomElementFactory(name: string): boolean;
	    /**
	     * Get the store with the given identifier.
	     *
	     * Note that the store may still need to be loaded when this method is called.
	     *
	     * @param id Identifier for the store
	     * @return A promise for when the store has been loaded. Rejected if loading fails or if no store is registered
	     *   with the given identifier.
	     */
	    getStore(id: Identifier): Promise<StoreLike>;
	    /**
	     * Check whether a store has been registered with the given identifier.
	     *
	     * @param id Identifier for the store
	     * @return `true` if a store has been registered, `false` otherwise.
	     */
	    hasStore(id: Identifier): boolean;
	    /**
	     * Get the widget with the given identifier.
	     *
	     * Note that the widget may still need to be loaded when this method is called.
	     *
	     * @param id Identifier for the widget
	     * @return A promise for when the widget has been loaded. Rejected if loading fails or if no widget is registered
	     *   with the given identifier.
	     */
	    getWidget(id: Identifier): Promise<WidgetLike>;
	    /**
	     * Check whether a widget has been registered with the given identifier.
	     *
	     * @param id Identifier for the widget
	     * @return `true` if a widget has been registered, `false` otherwise.
	     */
	    hasWidget(id: Identifier): boolean;
	}
	export interface AppMixin {
	    defaultStore?: StoreLike;
	    /**
	     * Register an action with the app.
	     *
	     * @param id How the action is identified
	     * @param action The action to be registered
	     * @return A handle to deregister the action
	     */
	    registerAction(id: Identifier, action: ActionLike): Handle;
	    /**
	     * Register an action factory with the app.
	     *
	     * The factory will be called the first time the action is needed. It'll be called with *one* argument:
	     * the combined registries of the app.
	     *
	     * Note that the `createAction()` factory from `dojo-actions` cannot be used here since it requires you to define
	     * the `do()` implementation, which the app factory does not allow.
	     *
	     * @param id How the action is identified
	     * @param factory A factory function that (asynchronously) creates an action.
	     * @return A handle to deregister the action factory, or the action itself once it's been created
	     */
	    registerActionFactory(id: Identifier, factory: ActionFactory): Handle;
	    /**
	     * Register a widget factory for a custom element.
	     *
	     * The factory will be called each time a widget instance is needed. It may be called with an options argument
	     * derived from a `data-options` attribute on the custom element that is to be be replaced by the created widget.
	     *
	     * @param name The name of the custom element. Must be valid according to
	     *   <https://www.w3.org/TR/custom-elements/#valid-custom-element-name>.
	     * @param factory A factory that (asynchronously) creates a widget.
	     * @return A handle to deregister the custom element
	     */
	    registerCustomElementFactory(name: string, factory: WidgetFactory): Handle;
	    /**
	     * Register a store with the app.
	     *
	     * @param id How the store is identified
	     * @param store The store to be registered
	     * @return A handle to deregister the store
	     */
	    registerStore(id: Identifier, store: StoreLike): Handle;
	    /**
	     * Register a store factory with the app.
	     *
	     * The factory will be called the first time the store is needed. It'll be called *without* any arguments.
	     *
	     * @param id How the store is identified
	     * @param factory A factory function that (asynchronously) creates a store.
	     * @return A handle to deregister the store factory, or the store itself once it's been created
	     */
	    registerStoreFactory(id: Identifier, factory: StoreFactory): Handle;
	    /**
	     * Register a widget with the app.
	     *
	     * @param id How the widget is identified
	     * @param widget The widget to be registered
	     * @return A handle to deregister the widget
	     */
	    registerWidget(id: Identifier, widget: WidgetLike): Handle;
	    /**
	     * Register a widget factory with the app.
	     *
	     * The factory will be called the first time the widget is needed. It'll be called with an options object
	     * that has its `id` property set to the widget ID.
	     *
	     * @param id How the widget is identified
	     * @param factory A factory function that (asynchronously) creates a widget.
	     * @return A handle to deregister the widget factory, or the widget itself once it's been created
	     */
	    registerWidgetFactory(id: Identifier, factory: WidgetFactory): Handle;
	    /**
	     * Load a POJO definition containing actions, stores and widgets that need to be registered.
	     *
	     * Action factories will be called with one argument: the combined registries of the app.
	     * Store and widget factories will also be called with one argument: an options object.
	     *
	     * @return A handle to deregister *all* actions, stores and widgets that were registered.
	     */
	    loadDefinition(definitions: Definitions): Handle;
	    /**
	     * Take a root element and replace <widget-instance> elements with widget instances.
	     *
	     * @param root The root element that is searched for <widget-instance> elements
	     * @return A handle to detach rendered widgets from the DOM
	     */
	    realize(root: Element): Promise<RealizationHandle>;
	    _registry: CombinedRegistry;
	    _resolveMid: ResolveMid;
	}
	export type App = AppMixin & CombinedRegistry;
	export interface AppOptions {
	    defaultStore?: StoreLike;
	    toAbsMid?: ToAbsMid;
	}
	export interface AppFactory extends ComposeFactory<App, AppOptions> {
	} const createApp: AppFactory;
	export default createApp;

}
declare module 'dojo-app/_factories' {
	import { ActionDefinition, ActionFactory, CombinedRegistry, CustomElementDefinition, StoreDefinition, StoreFactory, WidgetDefinition, WidgetFactory } from 'dojo-app/createApp';
	import { ResolveMid } from 'dojo-app/_moduleResolver';
	export function makeActionFactory(definition: ActionDefinition, resolveMid: ResolveMid): ActionFactory;
	export function makeCustomElementFactory(definition: CustomElementDefinition, resolveMid: ResolveMid): WidgetFactory;
	export function makeStoreFactory(definition: StoreDefinition, resolveMid: ResolveMid): StoreFactory;
	export function makeWidgetFactory(definition: WidgetDefinition, resolveMid: ResolveMid, registry: CombinedRegistry): WidgetFactory;

}
declare module 'dojo-app/main' {
	import createApp from 'dojo-app/createApp';
	export { createApp };

}
